#!/usr/bin/env perl
#
# Author: Tomi Ollila -- tomi.ollila t iki fi
#
#	Copyright (c) 2004 Tomi Ollila
#	    All rights reserved
#
# Created: Sun Sep 05 11:12:24 EEST 2004 too
# Last modified: Tue Nov 02 19:05:07 EET 2004 too
#
# This program is licensed under the GPL v2. See file COPYING for details.

use strict;
use warnings;

$" = ',';


#c# $0 = shift @ARGV;
#c# wait;
#c# my $m2vmp2cutbin = $0;
#c# my $m2vmp2cutbin2 = ($0 =~ m|^[^/].*/|)? "../$0": $0;

#c# my $mp2cutpoints = "$m2vmp2cutbin builtin mp2cutpoints.py";
#c# my $somehdrinfo  = "\$m2vmp2cutbin builtin mpg_somehdrinfo.py";
#c# my $catfiltered  = "\$m2vmp2cutbin builtin mpg_catfiltered.py";
#c# my $lvev6frames  = "$m2vmp2cutbin builtin lvev6frames.pl";

#s# SkipWhenEmbedded #s#

# Lines up to End Skip are deleted and s/#c# // done when this program
# is embedded to m2vmp2cut binary.

system 'cd src && exec make ../m2vmp2cut.mini';

my $toolpath = ($0 =~ m|^(/.*/?)|)? $1: '.';
my $toolpath2 = ($toolpath eq '.')? '..': $toolpath;

my $m2vmp2cutbin = "$toolpath/m2vmp2cut.mini";
my $m2vmp2cutbin2 = "$toolpath2/m2vmp2cut.mini";

my $mp2cutpoints = "$toolpath/mp2cutpoints.py";
my $somehdrinfo  = "$toolpath2/mpg_somehdrinfo.py";
my $catfiltered  = "$toolpath2/mpg_catfiltered.py";
my $lvev6frames  = "$toolpath/lvev6frames.pl";

undef $toolpath;
#s# End Skip ###

my $m2vscan   = "$m2vmp2cutbin builtin m2vscan";
my $mpgfilter = "\$m2vmp2cutbin builtin mpgfilter";

sub showdoc
{
    if ($_[0] != 0)
    {
	open STDOUT, ">&STDERR" || warn "1>&2 redirection failed: $! !!!\n";
    }

    system $m2vmp2cutbin ($0, 'builtin', 'showdoc', $_[1]);

    exit $_[0];
}

#print $0, " <$toolpath> \n";

my $asr = -1;
my $nomux = 0;
my $stop = 0;
my $sync = 0;
my $test = 0;
my $last = 0;
my $requant = 0.0;
my $mplexopts = '-M -f 8';

while (defined $ARGV[0] && $ARGV[0] =~ /^--/)
{
    $_ = $ARGV[0];
    $asr = $1 + 0, shift @ARGV, next if (/^--asr=(\d)\b/);
    $stop = 1, shift @ARGV, next if (/^--stop\b/);
    $nomux = 1, shift @ARGV, next if (/^--nomux\b/);
    $sync = $1 + 0, shift @ARGV, next if (/^--sync=(-?\d+)\b/);
    $test = $1 + 0, shift @ARGV, next if (/^--test=(\d+)\b/);
    $last = $1 + 0, shift @ARGV, next if (/^--last=(\d+)\b/);
    $requant = $1 + 0.0, shift @ARGV, next if (/^--requant=([1-5].\d+)\b/);
    $mplexopts = $1, shift @ARGV, next if ($ARGV[0] =~ /^--mplexopts=(.*)/);
    showdoc 0, 'Options' if (/^--options\b/);
    showdoc 0, 'Examples' if (/^--examples\b/);

    my $rv = 0;
    $rv = 1, print STDERR
      "\n$0: $ARGV[0]: unknown or incorrectly used option.\n"
	unless ($ARGV[0] =~ /^--help\b/);
    showdoc $rv, 'Usage';
}

my $vfile;
my $afile;
my $sfile;

if (@ARGV < 1 || ($ARGV[0] ne '+' && $ARGV[0] ne '-'))
{
    showdoc 1, 'Usage' if (@ARGV < 3);

    $vfile = $ARGV[1];
    $afile = $ARGV[2];
    $sfile = $ARGV[3]  if (defined $ARGV[3]);
}

if ($ARGV[0] eq '-')
{
    print "\nRunning $lvev6frames\n";
    exec "$lvev6frames";
}

#undef $sfile unless ($sfile);

sub basename
{
    local $_ = $_[0];
    s|.*/||; return $_;
}

my $framelist;
if ($ARGV[0] eq '+' || $ARGV[0] eq '--lve')
{
    chomp($_ = qx($lvev6frames));

    /(.*) (.*)/; my $lvevfile = $1; $framelist = $2;

    #print "--- video file $lvevfile, frames $framelist\n";

    exit 1 unless $2;
    # Usual case sanity-check -- does not take directories into account.

    my $x = basename($lvevfile);

    if ($ARGV[0] eq '+')
    {
	$vfile = $afile = $lvevfile;
	$afile =~ s/\.[^.]*$//;
	if (-f "$afile.mp2") { $afile = "$afile.mp2"; }
	elsif (-f "$afile.mpa") { $afile = "$afile.mpa"; }
	else
	{
	    print STDERR "$afile.mp2 not found.\n";
	    print STDERR "$afile.mpa not found.\n";
	    print STDERR "Cannot find audio file for $vfile.\n";
	    exit 1;
	}
	print "Using video file $vfile.\n";
	print "Using audio file $afile.\n";
    }
    else
    {
	die "lve file $lvevfile is not the same as in arg $vfile\n"
	  if system('test', $lvevfile, '-ef', $vfile) != 0;
    }
}
else
{
    $framelist = $ARGV[0];
    if ($framelist !~ /^(\d+-\d+)(,\d+-\d+)*$/)
    {
	warn "Incorrect frame range format in arg `$framelist'.\n";
	die "\t frame range format: nnn-nnn[,nnn-nnn]*\n";
    }
}

-r $vfile || die "Video file $vfile not readable.\n";
-r $afile || die "Audio file $afile not readable.\n";
-r $sfile || die "Subtitle file $sfile not readable.\n" if ($sfile);

if ($last)
{
    my @tlist;

    foreach (split(',', $framelist))
    {
	my ($s, $e) = split('-');
	last if ($s + 0 >= $last);
	$e = $last if ($last < $e + 0);

	push(@tlist, "$s-$e");
    }
    $framelist = "@tlist";
}

if ($test)
{
    if ($test > 0) { $test =  24 if ($test <  24); }
    else	   { $test = -24 if ($test > -24); }

    my @tlist;
    foreach (split(',', $framelist))
    {
	my ($s, $e) = split('-');
	push(@tlist,
	     sprintf("%d-%d,%d-%d", $s, $s + $test, $e - $test, $e));
    }
    $framelist = "@tlist";
}

my @ranges;
my $pe = -1;
my $frames = 0;

foreach (split(',', $framelist))
{
    my ($s, $e) = split('-');

    $frames += $e - $s;

    next if ($s == $e);
    die "$s greater than $e in a range in $ARGV[0].\n" if ($s > $e);

    die "\nRange begin $s is less than previous range end $pe in $ARGV[0].\n",
      "Which is not currently supported (as usually this is not desired).\n\n"
	if ($s < $pe);

    if ($s == $pe)
    {
	my $pr = pop(@ranges);
	$pr =~ s/-.*/-$e/;
	push @ranges, $pr;
    }
    else { push @ranges, "$s-$e"; }

    $pe = $e;
}

undef $framelist;
undef $pe;

my $pwd = `pwd` || die "Can not determine current directory.\n";

#
# Resolving suitable destination directory from source file names.
#

sub mtime
{
    my @l = stat $_[0];
    die "Could not stat $_[0]: $!\n" unless @l;
    return $l[9];
}

my $maxtime = mtime $vfile;
my $t = mtime $afile;
$maxtime = $t if ($t > $maxtime);
undef $t;

sub xtstamp
{
    my ($sec, $min, $hour, $mday, $mon, $year, undef, undef, undef)
      = localtime($_[0]);

    return sprintf("%d%s%02d+%02d%02d%02d",
		   $year % 10, substr('123456789ABC', $mon, 1), $mday,
		   $hour, $min, $sec);
}

my $vfile_bn = basename($vfile);

my $tstamp = xtstamp $maxtime;
my $odir = "$vfile_bn"; $odir =~ s/^(.*)\..*/$1-$tstamp/;
undef $tstamp;

#my $odir = "m2vmp2cut-dev";
#my $odir = "m2vmp2cut-$$";

unlink ',,,';
unless (-d $odir)
{
 print "\nCreating output directory $odir.\n";
 mkdir $odir, 0755;
}
else
{
    print "\nOutput directory $odir already exists.\n";
    system $m2vmp2cutbin, 'builtin', 'filerotate',
      $odir, 'audio.info', 'm2vmp2cut.log', 'do.sh';
}

open STDERR, "| tee $odir/m2vmp2cut.log"
  || die "STDERR redirection failed !!!: $! !!!\n";

open STDIN, ">&STDOUT" || die "0>&1 redirection failed: $! !!!\n";

print STDERR "--- Opened $odir/m2vmp2cut.log for output.\n";
#  "--- Note that do.sh logs here only when run via this driver.\n";

unless (-f "$odir/video.scan")
{
    print STDERR "Scanning video file $vfile.\n";
    exit 1 if system("$m2vscan $vfile $odir/video.scan") != 0;
}
else
{
    print STDERR "Video file $vfile already scanned (scan skipped).\n";
}

open I, "$odir/video.scan"
  || die "Can not open the scan file just created: $!\n";

# find last line from scan output file -- it contains average bit rate.

seek(I, -200, 2) || die "Seek failed: $!.\n";
my $abr;
{
    my $l;
    while (<I>)
    {
	$l = $_;
    }
    ($l =~ /abr\s+(\d+)/) || die "Incorrect scan file format.\n";
    $abr = int($1 / 1000);
}
seek(I, 0, 0) || die "Seek failed: $!.\n";


open SH, ">$odir/do.sh" || die "Can not create job script: $!\n";
chmod 0755, "$odir/do.sh";

shbegin();

print SH "# average video bit rate: $abr\n";
print SH "# video frames: @ranges\n";

my ($prevgopff, $prevgoppos,
    $currgopff, $currgoppos,
    $nextgopff, $nextgoppos) = (0,0,0,0,0,0);

my @asrs = (0,0,0,0,0,0,0,0,0,0);
sub lookx
{
    my $frame = $_[0];
    my $getasr = $_[1];

    return if ($nextgopff >= $frame);

    while (<I>)
    {
	# Currently we can manage with information up to frame number.
	next unless /^\s*(\d+)\s+\d+\s+(\d+)\s+(\d)\s/;

	($prevgopff, $prevgoppos) = ($currgopff, $currgoppos);
	($currgopff, $currgoppos) = ($nextgopff, $nextgoppos);
	($nextgopff, $nextgoppos) = ($2, $1);

	$asrs[$3] = $asrs[$3] + 1 if ($getasr);
	
	if ($2 >= $frame)
	{
	    return;
	}
    }
}

$vfile = '../' . $vfile unless ($vfile =~ m|/|);
print SH "\nvfile=$vfile\n";

my ($copystart, $copyend) = (0,0);

my ($tccopystart, $tccopyend, $tcframerange, $tcframeoffset) = (0,0,0,0);

my $tcfc = 1;

sub begincut
{
    my $frame = $_[0];

    return if ($frame == 0); # copystart = 0

    $tccopystart = $prevgoppos;
    $tccopyend = $nextgoppos ;# + 100 * 1024;

    my $e = $nextgopff - $prevgopff + $tcframeoffset;
    my $s = $frame - $prevgopff + $tcframeoffset;

    if ($tcframerange)
    {
	$tcframerange = "$tcframerange,$s-$e";
    }
    else { $tcframerange = "$s-$e"; }

    $copystart = $nextgoppos;

    #print SH " : +\n";
    print SH " \$m2vmp2cutbin builtin fileparts -q $tccopystart-$tccopyend";
    print SH " \$vfile >> tctmp.in.$tcfc\n";

    print SH " echo -n 001e | tr '01e' '\\000\\001\\267'",
      " >> tctmp.in.$tcfc\n";

    tcodecmdline($tcframerange);
}

sub endcut
{
    my $frame = $_[0];

    if ($currgoppos > $copystart)
    {
	print SH " : +\n";
	print SH " \$m2vmp2cutbin builtin fileparts -q ",
	  "$copystart-$currgoppos \$vfile\n\n";
    }

    $tccopystart = $prevgoppos;
    $tccopyend = $nextgoppos ; # + 100 * 1024;

    my $e = $frame - $prevgopff;
    my $s = $currgopff - $prevgopff;
    $tcframerange = "$s-$e";

    $copyend = $prevgoppos;
    $tcframeoffset = $nextgopff - $prevgopff;

    print SH " : +\n";
    print SH " \$m2vmp2cutbin builtin fileparts -q $tccopystart-$tccopyend";
    print SH " \$vfile > tctmp.in.$tcfc\n";

}

print SH "\n(\n /bin/rm -f tctmp.*\n";
my @timecodes;
print STDERR "Frame ranges: @ranges.\n";
foreach (@ranges)
{
    my ($s, $e) = split('-');

#    print "$s -- $e\n";

    lookx($s + 1, 0);
    print SH "# $prevgopff $currgopff $nextgopff ",
      "-- $prevgoppos $currgoppos $nextgoppos\n";
    begincut($s);

    lookx($e, 1);
    print SH "# $prevgopff $currgopff $nextgopff ",
      "-- $prevgoppos $currgoppos $nextgoppos\n";
    endcut($e);

    my ($ts, $te) = (palframe2timecode($s,$sync), palframe2timecode($e,$sync));

    push @timecodes, "$ts-$te";
}

print STDERR "Related timecodes: @timecodes\n";

my ($maxasrval, $maxasr, $counter) = (0, 0, -1);
print STDERR "Aspect ratios found (asr, count):\n";
foreach (@asrs)
{
    $counter++;
    next if ($_ == 0);
    print STDERR "  $counter, $_\n";
    if ($_ > $maxasrval)
    {
	$maxasr = $counter;
	$maxasrval = $_;
    }
}
print STDERR "Most common aspect: $maxasr ($maxasrval gops).\n";
if ($asr >= 0)
{
    print STDERR "... but using command line asr $asr.\n";
}
else { $asr = $maxasr; }

undef $maxasrval; undef $maxasr; undef $counter;


tcodecmdline($tcframerange);

my $reqq = '';
$reqq = "tcrequant -d 2 -f $requant |" if ($requant);

print SH ") | $reqq $mpgfilter $asr $frames ",
  $nomux? "incomplete.m2v": "fifo-video &", "\n";

undef $reqq;

sub palframe2timecode
{
    my $frametimems = $_[0] * 40 + $_[1];

    die "A/V sync correction requires adding silence audio, ",
      "which is currently not supported.\n" if ($frametimems < 0);

    my $ms = ($frametimems % 1000);
    my $s = int($frametimems / 1000) % 60;
    my $m = int($frametimems / 1000 / 60) % 60;
    my $h = int($frametimems / 1000 / 60 / 60);

    return sprintf "%02d:%02d:%02d.%03d", $h, $m, $s, $ms if ($h > 0);
    return sprintf "%02d:%02d.%03d", $m, $s, $ms if ($m > 0);
    return sprintf "%02d.%03d", $s, $ms;
}

#print "@timecodes", "\n";

print STDERR "Scanning audio for cut points.\n";
my $audiocutfilepos
  = `$mp2cutpoints "@timecodes" \"$afile\" \"$odir/audio.info\"`
  || exit 1;

$afile = '../' . $afile unless ($afile =~ m|/|);
print SH "\nafile=$afile\n";

chomp($audiocutfilepos);
print SH "\n# audio timecodes: @timecodes,\n#\tcutfilepos: $audiocutfilepos\n";

print SH "\n\$m2vmp2cutbin builtin fileparts $audiocutfilepos";
print SH " \$afile > ", $nomux? "incomplete.mp2": "fifo-audio &", "\n";


unless ($nomux)
{
    print SH "\nmplex $mplexopts -o incomplete.mpg fifo-video fifo-audio ",
      "> mplex.out 2>&1\n";
}
#
# Final lines.
#

print SH "\n", "wait\n", "trap - 0\n", ": '***' `date`\n", "set +x\n";

if ($nomux) { print SH "mv -f incomplete.m2v out.m2v\n",
		"mv -f incomplete.mp2 out.mp2\n"; }
else { print SH "mv -f incomplete.mpg out.mpg\n"; }

print SH ": `date`; ls -lrtac ../$odir\n";

close SH;

if ($stop)
{
    print "\nWork script $odir/do.sh\ncreated but not executed by request.\n";
    exit 0;
}

print STDERR "Running work script $odir/do.sh...\n";

#$SIG{'INT'} = 'sleep 1; exit 1';

#system 'script', '-c', "$odir/do.sh", "$odir/do.sh-out";
if (system("$odir/do.sh") != 0)
{
    print "\n$odir/do.sh was partly unsuccessful. exiting.\n";
    select undef, undef, undef, 0.25;
    exit 1;
}

  print "\nCreating link ,,, -> $odir/\n";
unlink ',,,';
symlink "$odir", ',,,';

print "\nDone, Result (if any) in $odir/",
  $nomux? "out.(m2v|mp2)": "out.mpg", ".\n\n";

# ----------------------- Shell script writing parts --

sub shbegin
{
    if ( -f '/bin/bash')
    {
	print SH "#!/bin/bash\n";
    }
    elsif ( -f '/bin/ksh')
    {
	print SH "#!/bin/ksh\n";
    }
    else
    {
	die "Can not find supported shell, neither /bin/bash nor /bin/ksh.\n";
    }

    print SH << "EOF";

# This program is licensed under the GPL v2. See file COPYING for details.

m2vmp2cutbin=$m2vmp2cutbin2

mainpid=\$\$
trap 'kill -USR1 0' 0

[ -f video.scan ] || cd $odir

\$m2vmp2cutbin builtin vermatch 2 || exit 1

EOF

    unless ($nomux)
    {
	print SH "rm -f fifo-video fifo-audio\n",
	  "mkfifo fifo-video fifo-audio\n";
    }
    print SH << "EOF";

exec 0>&1

\$m2vmp2cutbin builtin filerotate . transcode.out mplex.out
rm -f out.*

set -x; : '***' `date`

[ -f video.scan ] || exit 0

EOF
}

sub tcodecmdline
{
    my $range = $_[0];

    print SH << "EOF";

 set x `$somehdrinfo tctmp.in.$tcfc tctmp.matrix.$tcfc`
 [ x\$3 = x0 ] && np=--encode_fields=t || np=
 [ x\$5 = x1 -o x\$7 = x1 ] && fa='-K file=tctmp.matrix.$tcfc' || fa=

 echo >> transcode.out
 transcode -V -w $abr -x auto,null -y mpeg2enc,null -F "3,-R 2 \$fa" \$np \\
	-i tctmp.in.$tcfc -o tctmp.out.$tcfc -c $range >> transcode.out 2>&1

 $catfiltered \$5 \$7 \$9 tctmp.out.$tcfc.m2v

# transcode -V -w $abr -x auto,null -y ffmpeg,null -F mpeg2video \\
##	--export_asr 2 \\
#	-i tctmp.in.$tcfc -o tctmp.out.$tcfc \\
#	-c $range >> transcode.out 2>&1
#
# /bin/cat tctmp.out.$tcfc.m2v

EOF
 $tcfc++;
}
