/*
 * fileparts.c
 *
 * Author: Tomi Ollila too t iki fi
 *
 *	Copyright (c) 2004 Tomi Ollila
 *	    All rights reserved
 *
 * Created: Sun Sep 26 08:51:22 EEST 2004 too
 * Last modified: Tue Nov 02 19:04:17 EET 2004 too
 *
 * This program is licensed under the GPL v2. See file COPYING for details.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#define DBGS 0
#include "x.h"
#include "fileparts.h"
#include "ghdrs/fileparts_priv.h"

void builtin_fileparts(int argc, char ** argv)
{
  char * partstring;
  char * ifile;
  int fd;
  char * p;
  off_t tlen = 0;
  bool quiet = false;
  
  if (argc >= 5 && strcmp(argv[3], "-q") == 0)
    {
      quiet = true;
      argc--; argv++;
    }
  
  if (argc < 5)
    xerrf("argc < 5\n");

  partstring = argv[3];
  ifile = argv[4];

  if (! s__chkpartstring(partstring))
    xerrf("%s: incorrect string.\n", partstring);

#if 1
  if ((fd = open(ifile, O_RDONLY)) < 0)
    xerrf("open O_DRONLY failed:");
#endif
  p = partstring;
  while (true)
    {
      off_t start = atoll(p);
      off_t len;

#if 1
      if (lseek(fd, start, SEEK_SET) < 0)
	xerrf("lseek failed:");
#endif
      p = strchr(p, '-') + 1;
      len = atoll(p) - start;

#if 1
      if (s__fdcopy(fd, 1, len) != len)
	xerrf("Data copy failure:");	
#else
      printf("start %lld, len %lld\n", start, len);
#endif
      tlen += len;
      
      p = strchr(p, ',');
      if (p == NULL)
	break;
      p++;
    }

  if (! quiet)
    fprintf(stderr, "fileparts: Transferred %lld bytes of data.\n", tlen);
  
  exit(0);
}

static bool s__chkpartstring(const char * partstring)
{
  const char * p = partstring;
  char c;
  int state;
  int start, end; /* unavoidable warnings here -- but don't suppress those !!*/

  /* using regexp /^(\d+-\d+)(,\d+-\d+)*$/) would have been easier. */
  
  for (state = 0, c = *p;   c;   p++, c = *p)
    switch (state)
      {
      case 0: /* [0-9] */
	if (isdigit(c)) { start = atoi(p); state = 1; }
	else		return false;
	break;
	
      case 1: /* [0-9]*-? */
	if (isdigit(c))	continue;
	if (c == '-')	state = 2;
	else		return false;
	break;

      case 2: /* [0-9] */
	if (isdigit(c)) { end = atoi(p); state = 3; }
	else		return false;
	break;
	
      case 3:  /* [0-9]*,? */
	if (isdigit(c)) continue;
	if (c == ',')	{ if (start > end) return false; state = 0; }
	else		return false;
      }
  
  if (state != 3)  return false;
  if (start > end) return false;

  return true;
}
  
static off_t s__fdcopy(int in, int out, off_t len)
{
  unsigned char megabuf[1024 * 1024];
  off_t tlen = 0;
  ssize_t fl;
  
  for(; len > 0; )
    {
      int l = read(in, megabuf, len > sizeof megabuf? sizeof megabuf: len);

      if (l <= 0)
	{
	  if (errno == EINTR)
	    continue;
	  return tlen;
	}

      fl = writefully(out, megabuf, l);
      tlen += fl;
      
      if (fl != l)
	break;

      len -= l;
    }

  return tlen;
}


/*
 * Local variables:
 * mode: c
 * c-file-style: "gnu"
 * tab-width: 8
 * End:
 */
