/*
 * m2vscan.c
 *
 * Author: Tomi Ollila too t iki fi
 *
 *	Copyright (c) 2004 Tomi Ollila
 *	    All rights reserved
 *
 * Created: Sun Sep 26 13:23:29 EEST 2004 too
 * Last modified: Wed Oct 20 20:48:34 EEST 2004 too
 *
 * This program is licensed under the GPL v2. See file COPYING for details.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define DBGS 0
#include "x.h"
#include "mpgdata.h"
#include "bufwrite.h"

#include "m2vscan.h"
#include "ghdrs/m2vscan_priv.h"

void builtin_m2vscan(int argc, char ** argv)
{
  char * ifile;
  char * ofile;
  int ifd, ofd;
  struct stat st;
  
  if (argc < 5)
    xerrf("argc < 5\n");

  ifile = argv[3];
  ofile = argv[4];

  if ((ifd = open(ifile, O_RDONLY)) < 0)
    xerrf("open O_DRONLY failed:");

  if (fstat(ifd, &st) < 0)
    xerrf("fstat failed:");
  
  if ((ofd = open(ofile, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0)
    xerrf("open (for write) failed:");

  s__scan(ifd, ofd, st.st_size);
}

static void s__scan(int ifd, int ofd,
		    off_t fsize) /* protoadd GCCATTR_NORETURN */
{
  MpgBuf mpgbuf;
  unsigned char ibuf[65536];

  FILE * ofh = fdopen(ofd, "w");
  
  off_t filepos;
  int pictures = 0;
  int gops = 0;
  int goppic = 0;
  off_t seqpos = 0;
  int asr = 0;

  int prevpc = 0;
  
  if (ofh == NULL)
    xerrf("fdopen \"w\" failed:");

  mpgdata_init(&mpgbuf, ifd, ibuf, sizeof ibuf);
  
  for (filepos = 0; mpgdata(&mpgbuf); filepos += mpgbuf.len)
    {
      int cpc = filepos * 100 / fsize;

      if (cpc != prevpc)
	{
	  printf("\r- Scanning video at %lld of %lld (%d%%) ",
		 filepos, fsize, cpc);
	  fflush(stdout);
	  prevpc = cpc;
	}
      
      if (mpgbuf.len >= 3)
	{
	  unsigned char * data = mpgbuf.data; /*printf("%d\n", data[3]); */

	  switch (data[3])
	    {
	    case 0x00: /* picture header */
	      pictures++;
	      break;

	    case 0xb3:	/* sequence header */
	      seqpos = filepos;
	      if (mpgbuf.len > 7)
		{
		  int frame_rate = data[7] & 0x0f;

		  if (frame_rate != 3)
		    {
		      xerrf("\rCurrently we can only support PAL (25fps) "
			    "frame rate.\n");
		    }
		  asr = data[7] >> 4;
		}
	      break;

	    case 0xb8: /* group of pictures (GOP) header */
	      if (mpgbuf.len > 7)
		{
                  int drop_frame_flag = data[4] & 0x80;
                  int hour   = (data[4] & 0x7f) >> 2;
                  int minute = ((data[4] & 0x03) << 4) + (data[5] >> 4);
                  int second = ((data[5] & 0x07) << 3) + (data[6] >> 5);
                  int frame  = ((data[6] & 0x1f) << 1) + (data[7] >> 7);
                  int closed_gop_flag = data[7] & 0x40;
                  int broken_gop_flag = data[7] & 0x20;

		  int time = (((hour * 60 + minute) * 60 + second) * 25) + frame;
		  if (pictures)
		    fprintf(ofh, "%d\n", pictures - goppic);
		  else
		    fprintf(ofh, "#file offset gop# frame# asr _ time _   "
			    "c/o  # of frames ver1\n");
		  
		  if (time != pictures)
		    {
		      fprintf(stderr,"\nTimecode problem at filepos %lld..., ",
			      filepos);
		      fprintf(stderr, "%02d:%02d:%02d.%02d --",
			      hour, minute, second, frame);
		      fprintf(stderr, "%d %d\n", time, pictures);
		      exit(1);
		    }

		  goppic = pictures;
		  fprintf(ofh, "%10lld %5d %6d  %d  %02d:%02d:%02d.%02d  %c  ",
			  seqpos, gops, goppic, asr,
			  hour, minute, second, frame,
			  closed_gop_flag? 'c': 'o');
		  
		  gops++;
		}
	    }
	}
    }
  fprintf(ofh, "%d\n", pictures - goppic);
  fprintf(ofh, "end: size %lld frames %d gops %d abr %ld\n",
	  filepos, pictures, gops, (long)(filepos / (pictures / 25) * 8));
  printf("\r- Scanning video at %lld of %lld (100%%).\n", filepos, fsize);
  exit(0);
}


/*
 * Local variables:
 * mode: c
 * c-file-style: "gnu"
 * tab-width: 8
 * End:
 */
