/*
 * mpgfilter.c
 *
 * Author: Tomi Ollila too t iki fi
 *
 *	Copyright (c) 2004 Tomi Ollila
 *	    All rights reserved
 *
 * Created: Sat Sep 25 16:35:59 EEST 2004 too
 * Last modified: Tue Nov 02 18:26:27 EET 2004 too
 *
 * This program is licensed under the GPL v2. See file COPYING for details.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdlib.h>

#define DBGS 0
#include "x.h"
#include "mpgdata.h"
#include "bufwrite.h"

#include "mpgfilter.h"
#include "ghdrs/mpgfilter_priv.h"

void builtin_mpgfilter(int argc, char ** argv)
{
  MpgBuf mpgbuf;
  unsigned char ibuf[65536];

  BufWrite bw;
  unsigned char obuf[65536];
  
  int asr;
  int frames;
  char * ofile;
  int fd;
  
  if (argc < 6)
    xerrf("argc < 6\n");

  asr = atoi(argv[3]);
  frames = atoi(argv[4]);
  ofile = argv[5];

  if (frames <= 0)
    xerrf("frames <= 0 (%d)\n", frames);

  if (asr <= 0 && argv[3][0] != '0')
    xerrf("asr incorrect (%s)\n", argv[3]);
  
  if ((fd = open(ofile, O_WRONLY|O_CREAT|O_TRUNC, 0755)) < 0)
    xerrf("open (write) failed:");
  
  bufwrite_init(&bw, fd, obuf, sizeof obuf);
  mpgdata_init(&mpgbuf, 0, ibuf, sizeof ibuf);

  s__thefilter(&mpgbuf, &bw, asr, frames);
  
}

static void s__thefilter(MpgBuf * mpgbuf, BufWrite * bw,
			 int asr, int frames) /* protoadd GCCATTR_NORETURN */
{
  int t1;
  
  int fcount = 0;
  int prevpc = 0;

  off_t tlen = 0;
  
  /* Ignore unknown aspect ratios... */
  if (asr > 4)
    asr = 0;
  else /* tune it to be already suitable... to write in. */
    asr <<= 4;
  
  while (mpgdata(mpgbuf))
    {
      tlen += mpgbuf->len;
      if (mpgbuf->len >= 3)
	{
	  /*printf("%d\n", mpgbuf->data[3]); */
	  switch (mpgbuf->data[3])
	    {
	    case 0x00: /* picture header */
	      fcount++;
	      t1 = fcount * 100 / frames;
	      if (t1 != prevpc)
		{
		  printf("\r- Picture %d of %d total (%d%%) ",
			 fcount, frames, fcount * 100 / frames);
		  fflush(stdout);
		  prevpc = t1;
		}
	      break;
	    case 0xb3:	/* sequence header */
	      if (asr > 0)
		if (mpgbuf->len > 7)
		  mpgbuf->data[7] = (mpgbuf->data[7] & 0x0f) | asr;
	      break;
	    }
	}
      if (! bufwrite(bw, mpgbuf->data, mpgbuf->len))
	/* The value written below may not be exact */
	xerrf("mpg video data write failure after writing %lld bytes:", tlen);
    }

  printf("\n");
  if (! bufwrite_exit(bw))
    /* The value written below may not be exact */
    xerrf("mpg video data write failure after writing %lld bytes:", tlen);

  fprintf(stderr, "Transferred %lld bytes of mpeg video data.\n", tlen);
  
  exit(0);
}  
  

/*
 * Local variables:
 * mode: c
 * c-file-style: "gnu"
 * tab-width: 8
 * End:
 */
